@@ -46,15 +46,16 @@ class Bsm(opt.OptAnalyticABC):
             Vanilla option price
         """
         disc_fac = np.exp(-texp * intr)
-        fwd = spot * (1.0 if is_fwd else np.exp(-texp * divr) / disc_fac)
+        fwd = np.array(spot) * (1.0 if is_fwd else np.exp(-texp * divr) / disc_fac)
 
-        sigma_std = np.maximum(sigma * np.sqrt(texp), np.finfo(float).eps)
+        sigma_std = np.maximum(np.array(sigma) * np.sqrt(texp), np.finfo(float).eps)
 
         # don't directly compute d1 just in case sigma_std is infty
         d1 = np.log(fwd / strike) / sigma_std
         d2 = d1 - 0.5*sigma_std
         d1 += 0.5*sigma_std
 
+        cp = np.array(cp)
         price = fwd * spst.norm.cdf(cp * d1) - strike * spst.norm.cdf(cp * d2)
         price *= cp * disc_fac
         return price
@@ -80,6 +81,15 @@ class Bsm(opt.OptAnalyticABC):
         delta *= df if self.is_fwd else divf
         return delta
 
+    def cdf(self, strike, spot, texp, cp=1):
+
+        fwd, df, divf = self._fwd_factor(spot, texp)
+
+        sigma_std = np.maximum(self.sigma * np.sqrt(texp), np.finfo(float).eps)
+        d2 = np.log(fwd / strike) / sigma_std - 0.5 * sigma_std
+        cdf = spst.norm.cdf(cp * d2)  # formula according to wikipedia
+        return cdf
+
     def gamma(self, strike, spot, texp, cp=1):
 
         fwd, df, divf = self._fwd_factor(spot, texp)
@@ -356,6 +366,10 @@ class BsmDisp(smile.OptSmileABC):
         return self.bsm_model.delta(
             self.disp(strike), self.disp(spot), *args, **kwargs)
 
+    def cdf(self, strike, spot, *args, **kwargs):
+        return self.bsm_model.cdf(
+            self.disp(strike), self.disp(spot), *args, **kwargs)
+
     def vega(self, strike, spot, *args, **kwargs):
         return self.bsm_model.vega(
             self.disp(strike), self.disp(spot), *args, **kwargs)
