@@ -7,10 +7,10 @@ Created on Tue Oct 10
 
 import abc
 import copy
+import os
+import pandas as pd
 import numpy as np
 import scipy.optimize as spop
-#import scipy.stats as spst
-#import scipy.special as spsp
 
 from . import opt_smile_abc as smile
 from . import bsm
@@ -28,7 +28,7 @@ class SabrABC(smile.OptSmileABC, abc.ABC):
             sigma: model volatility at t=0
             vov: volatility of volatility
             rho: correlation between price and volatility
-            beta: elasticity parameter. 0.5 by default
+            beta: elasticity parameter. 1.0 by default
             intr: interest rate (domestic interest rate)
             divr: dividend/convenience yield (foreign interest rate)
             is_fwd: if True, treat `spot` as forward price. False by default.
@@ -45,7 +45,7 @@ class SabrABC(smile.OptSmileABC, abc.ABC):
 
     def _variables(self, fwd, texp):
         betac = 1.0 - self.beta
-        alpha = self.sigma / np.power(fwd, betac) if self.beta > 0.0 else self.sigma
+        alpha = self.sigma / np.power(fwd, betac) #if self.beta > 0.0 else self.sigma
         rho2 = self.rho * self.rho
         rhoc = np.sqrt(1.0 - rho2)
         vovn = self.vov * np.sqrt(np.maximum(texp, 1e-64))
@@ -53,28 +53,79 @@ class SabrABC(smile.OptSmileABC, abc.ABC):
 
     def _m_base(self, vol):
         """
-        The model of the vol_for_price
+        Create base model based on _base_beta value: Norm for 0, Cev for (0,1), and Bsm for 1
+        If _base_beta is None, use base instead.
 
         Args:
-            vol:
+            vol: base model volatility
 
         Returns: model
         """
-        vol_beta = self.beta if self._base_beta is None else self._base_beta
+        base_beta = self._base_beta or self.beta
 
-        if abs(vol_beta - 1.0) < 0.001:
+        if np.isclose(base_beta, 1):
             return bsm.Bsm(vol, intr=self.intr, divr=self.divr)
-        elif abs(vol_beta) < 0.001:
+        elif np.isclose(base_beta, 0):
             return norm.Norm(vol, intr=self.intr, divr=self.divr)
         else:
-            return cev.Cev(vol, beta=vol_beta, intr=self.intr, divr=self.divr)
+            return cev.Cev(vol, beta=base_beta, intr=self.intr, divr=self.divr)
 
     def vol_smile(self, strike, spot, texp, cp=1, model=None):
         if model is None:
-            model = 'bsm' if self.beta > 0.0 else 'norm' if self.beta == 0 else None
+            model = 'norm' if np.isclose(self.beta, 0) else 'bsm' if self.beta > 0.0 else None
 
         return super().vol_smile(strike, spot, texp, cp=1, model=model)
 
+    @classmethod
+    def init_benchmark(cls, set_no=None):
+        """
+        Initiate a SABR model with stored benchmark parameter sets
+
+        Args:
+            set_no: set number
+
+        Returns:
+            Dataframe of all test cases if set_no = None
+            (model, Dataframe of result, params) if set_no is specified
+
+        References:
+            Antonov, Alexander, Konikov, M., & Spector, M. (2013). SABR spreads its wings. Risk, 2013(Aug), 58–63.
+            Antonov, Alexandre, Konikov, M., & Spector, M. (2019). Modern SABR Analytics. Springer International Publishing. https://doi.org/10.1007/978-3-030-10656-0
+            Antonov, Alexandre, & Spector, M. (2012). Advanced analytics for the SABR model. Available at SSRN. https://ssrn.com/abstract=2026350
+            Cai, N., Song, Y., & Chen, N. (2017). Exact Simulation of the SABR Model. Operations Research, 65(4), 931–951. https://doi.org/10.1287/opre.2017.1617
+            Korn, R., & Tang, S. (2013). Exact analytical solution for the normal SABR model. Wilmott Magazine, 2013(7), 64–69. https://doi.org/10.1002/wilm.10235
+            Lewis, A. L. (2016). Option valuation under stochastic volatility II: With Mathematica code. Finance Press.
+            von Sydow, L., ..., Haentjens, T., & Waldén, J. (2018). BENCHOP - SLV: The BENCHmarking project in Option Pricing – Stochastic and Local Volatility problems. International Journal of Computer Mathematics, 1–14. https://doi.org/10.1080/00207160.2018.1544368
+        """
+        this_dir, _ = os.path.split(__file__)
+        file = os.path.join(this_dir, 'data/sabr_benchmark.xlsx')
+        df_param = pd.read_excel(file, sheet_name='Param', index_col='Sheet')
+
+        if set_no is None:
+            return df_param
+        else:
+            df_val = pd.read_excel(file, sheet_name=str(set_no))
+            param = df_param.loc[set_no]
+            args_model = {k: param[k] for k in ('sigma', 'vov', 'rho', 'beta')}
+            args_pricing = {k: param[k] for k in ('texp', 'spot')}
+
+            assert(df_val.columns[0] == 'k' or df_val.columns[0] == 'K')
+            args_pricing['strike'] = df_val.values[:, 0]
+            if df_val.columns[0] == 'k':
+                args_pricing['strike'] *= param['spot']
+
+            val = df_val[param['col_name']].values
+            is_iv = param['col_name'].startswith('IV')
+
+            m = cls(**args_model)
+
+            param_dict = {
+                'args_pricing': args_pricing,
+                'ref': param['Reference'], 'val': val, 'is_iv': is_iv
+            }
+
+            return m, df_val, param_dict
+
 
 class SabrVolApproxABC(SabrABC):
     """
@@ -132,7 +183,7 @@ class SabrVolApproxABC(SabrABC):
         """
         rho2 = rho * rho
         # initalization with expansion for for small |zz|
-        xx_zz = 1 - zz * ((rho / 2) - zz * ((1 / 2 * rho2 - 1 / 6) - 1 / 8 * (5 * rho2 - 3) * rho * zz))
+        xx_zz = 1 - (zz/2)*(rho - zz*(rho2 - 1/3 - (5*rho2 - 3)/4*rho*zz))
 
         yy = SabrVolApproxABC._vv(zz, rho)
         eps = 1e-5
@@ -169,7 +220,7 @@ class SabrVolApproxABC(SabrABC):
     def impvol(self, price, strike, spot, texp, cp=1, setval=False):
         model = copy.copy(self)
 
-        vol = self._m_base().impvol(price, strike, spot, texp, cp=cp)
+        vol = self._m_base(None).impvol(price, strike, spot, texp, cp=cp)
 
         def iv_func(_sigma):
             model.sigma = _sigma
@@ -181,15 +232,16 @@ class SabrVolApproxABC(SabrABC):
             self.sigma = sigma
         return sigma
 
-    def vol_smile(self, strike, spot, texp, model=None, cp=1):
+    def vol_smile(self, strike, spot, texp, cp=1, model=None):
         if model is None:
-            model = 'bsm' if self.beta > 0.0 else 'norm' if self.beta == 0 else None
+            model = 'norm' if np.isclose(self.beta, 0) else 'bsm' if self.beta > 0.0 else None
 
-        vol_beta = self.beta if self._base_beta is None else self._base_beta
-        if (model.lower() == 'bsm' and vol_beta == 1.0) or (model.lower() == 'norm' and vol_beta == 0.0):
+        vol_beta = self._base_beta or self.beta
+        if (model.lower() == 'bsm' and np.isclose(vol_beta, 1)) or \
+                (model.lower() == 'norm' and np.isclose(vol_beta, 0)):
             vol = self.vol_for_price(strike, spot, texp)
         else:
-            vol = super().vol_smile(strike, spot, texp, model=model, cp=cp)
+            vol = super().vol_smile(strike, spot, texp, cp=cp, model=model)
         return vol
 
 
@@ -218,31 +270,31 @@ class SabrHagan2002(SabrVolApproxABC):
 
         if texp <= 0.0:
             return 0.0
-
         fwd, _, _ = self._fwd_factor(spot, texp)
-        _alpha, betac, rhoc, rho2, vovn = self._variables(spot, texp)
+        alpha, betac, rhoc, rho2, vovn = self._variables(spot, texp)
         betac2 = betac**2
 
-        #kk = strike / fwd  # standardized strike
+        log_kk = np.log(fwd/strike)
+        log_kk2 = log_kk*log_kk
+        pow_kk = np.power(strike/fwd, betac/2)
 
-        powFwdStrk = np.power(fwd*strike, betac/2)
-        logFwdStrk = np.log(fwd/strike)
-        logFwdStrk2 = logFwdStrk*logFwdStrk
+        pre1 = pow_kk*(1 + betac2/24*log_kk2*(1 + betac2/80*log_kk2))
 
-        pre1 = powFwdStrk*(1 + betac2/24*logFwdStrk2*(1 + betac2/80*logFwdStrk2))
+        term02 = (2 - 3*rho2)/24 * self.vov**2
+        term11 = alpha*self.vov*self.rho*self.beta/4/pow_kk
+        term20 = (betac*alpha/pow_kk)**2 / 24
 
-        pre2alp0 = (2 - 3*rho2)*self.vov*self.vov/24
-        pre2alp1 = self.vov*self.rho*self.beta/4/powFwdStrk
-        pre2alp2 = betac2/24/(powFwdStrk*powFwdStrk)
+        if self.approx_order == 0:
+            vol = 1.0
+        else:
+            vol = 1.0 + texp*(term02 + term11 + term20)
 
-        pre2 = 1.0 + texp*(pre2alp0 + self.sigma*(pre2alp1 + pre2alp2*self.sigma))
+        zz = pow_kk*log_kk*self.vov/np.maximum(alpha, np.finfo(float).eps)
+        hh = self._hh(-zz, self.rho)  # note we pass -zz becaues hh(zz) definition is different
+        vol *= alpha*hh/pre1   # bsm vol
+        return vol
 
-        zz = powFwdStrk*logFwdStrk*self.vov/np.maximum(self.sigma, np.finfo(float).eps)  # need to make sure sig > 0
-        HH = self._hh(-zz, self.rho)  # note we pass -zz becaues hh(zz) definition is different
-        bsmvol = self.sigma*HH*pre2/pre1   # bsm vol
-        return bsmvol
-
-    def calibrate3(self, price_or_vol3, strike3, spot, texp=None, cp=1, setval=False, is_vol=True):
+    def calibrate3(self, price_or_vol3, strike3, spot, texp, cp=1, setval=False, is_vol=True):
         """
         Given option prices or normal vols at 3 strikes, compute the sigma, vov, rho to fit the data
         If prices are given (is_vol=False) convert the prices to vol first.
@@ -255,15 +307,16 @@ class SabrHagan2002(SabrVolApproxABC):
         if is_vol:
             vol3 = price_or_vol3
         else:
-            vol3 = self._m_base().impvol(price_or_vol3, strike3, spot, texp)
+            vol3 = self._m_base(None).impvol(price_or_vol3, strike3, spot, texp, cp=cp)
 
         def iv_func(x):
             model.sigma = np.exp(x[0])
             model.vov = np.exp(x[1])
             model.rho = np.tanh(x[2])
-            return model.vol_for_price(strike3, spot, texp=texp) - vol3
+            err = model.vol_for_price(strike3, spot, texp=texp) - vol3
+            return err
 
-        sol = spop.root(iv_func, np.array([vol3[1], -1, 0.0]))
+        sol = spop.root(iv_func, np.array([np.log(vol3[1]), -1, 0.0]))
         params = {"sigma": np.exp(sol.x[0]), "vov": np.exp(sol.x[1]), "rho": np.tanh(sol.x[2])}
 
         if setval:
@@ -271,7 +324,55 @@ class SabrHagan2002(SabrVolApproxABC):
 
         return params
 
-    
+
+class SabrNorm(SabrVolApproxABC):
+    """
+    Noram SABR model (beta=0) with normal volatility approximation.
+
+    Examples:
+        >>> import numpy as np
+        >>> import pyfeng as pf
+        >>> m = pf.SabrNorm(sigma=2, vov=0.2, rho=-0.3, beta=0.5)
+        >>> m.vol_for_price(np.arange(80, 121, 10), 100, 1.2)
+        array([0.21976016, 0.20922027, 0.200432  , 0.19311113, 0.18703486])
+        >>> m.price(np.arange(80, 121, 10), 100, 1.2)
+        array([22.04862858, 14.56226187,  8.74170415,  4.72352155,  2.28891776])
+    """
+    _base_beta = 0  # should not be changed
+    is_atmvol = False
+
+    def __init__(self, sigma, vov=0.0, rho=0.0, beta=None, intr=0.0, divr=0.0, is_fwd=False, is_atmvol=False):
+        """
+        Args:
+            sigma: model volatility at t=0
+            vov: volatility of volatility
+            rho: correlation between price and volatility
+            beta: elasticity parameter. should be 0 or None.
+            intr: interest rate (domestic interest rate)
+            divr: dividend/convenience yield (foreign interest rate)
+            is_fwd: if True, treat `spot` as forward price. False by default.
+            is_atmvol: If True, use `sigma` as the ATM normal vol
+        """
+        # Make sure beta = 0
+        if beta is not None and not np.isclose(beta, 0.0):
+            print(f'Ignoring beta = {beta}...')
+        self.is_atmvol = is_atmvol
+        super().__init__(sigma, vov, rho, beta=0, intr=intr, divr=divr, is_fwd=is_fwd)
+
+    def vol_for_price(self, strike, spot, texp):
+        fwd = self.forward(spot, texp)
+
+        if self.approx_order == 0 or self.is_atmvol:
+            vol = 1.0
+        else:
+            vol = 1.0 + texp*(2 - 3*self.rho**2)/24 * self.vov**2
+
+        zz = self.vov/self.sigma*(strike-fwd)
+        hh = self._hh(zz, self.rho)
+        vol *= self.sigma * hh
+        return vol
+
+
 class SabrChoiWu2021H(SabrVolApproxABC, smile.MassZeroABC):
     """
     The CEV volatility approximation of the SABR modelbased on Theorem 1 of Choi & Wu (2019)
@@ -296,20 +397,20 @@ class SabrChoiWu2021H(SabrVolApproxABC, smile.MassZeroABC):
             sigma: model volatility at t=0
             vov: volatility of volatility
             rho: correlation between price and volatility
-            beta: elasticity parameter. 0.5 by default
+            beta: elasticity parameter. 1.0 by default
             intr: interest rate (domestic interest rate)
             divr: dividend/convenience yield (foreign interest rate)
             is_fwd: if True, treat `spot` as forward price. False by default.
             vol_beta: the beta for the volatility to choose _m_vol. If None (by default) vol_beta = beta
         """
-        self.vol_beta = vol_beta
+        self._base_beta = vol_beta
         super().__init__(sigma, vov, rho, beta, intr, divr, is_fwd)
 
     def vol_for_price(self, strike, spot, texp):
         # fwd, spot, sigma may be either scalar or np.array.
         # texp, vov, rho, beta should be scholar values
 
-        vol_beta = self.beta if self.vol_beta is None else self.vol_beta
+        vol_beta = self._base_beta or self.beta
         vol_betac = 1.0 - vol_beta
 
         fwd, _, _ = self._fwd_factor(spot, texp)
@@ -319,7 +420,7 @@ class SabrChoiWu2021H(SabrVolApproxABC, smile.MassZeroABC):
 
         vov_over_alpha_safe = self.vov/np.maximum(alpha, np.finfo(float).eps)
         tmp = self._int_inv_locvol(kk, self.beta)
-        qq_ratio = 1.0 if self.vol_beta is None else self._int_inv_locvol(kk, vol_beta) / tmp
+        qq_ratio = 1.0 if self._base_beta is None else self._int_inv_locvol(kk, vol_beta) / tmp
 
         qq = tmp * (kk-1.0)
         zz = vov_over_alpha_safe * qq  # zeta = (vov/sigma0) q
@@ -332,7 +433,7 @@ class SabrChoiWu2021H(SabrVolApproxABC, smile.MassZeroABC):
         term11 = self.rho*self.beta/4 * self.vov*alpha * self._int_inv_locvol(kk, betac)
 
         # term20: O(alpha^2)
-        if self.vol_beta is None or self.beta == vol_beta:
+        if np.isclose(self.beta, vol_beta):
             term20 = 0.0
         else:
             with np.errstate(divide='ignore', invalid='ignore'):
@@ -340,29 +441,26 @@ class SabrChoiWu2021H(SabrVolApproxABC, smile.MassZeroABC):
                 term20 = np.where(
                     np.fabs(qq) < 1e-6,
                     (np.square(betac) - np.square(vol_betac)) / 24 * np.square(alpha),
-                    (0.5*(self.beta-self.vol_beta)*np.log(kk) - np.log(qq_ratio)) * np.square(alpha)/qq
+                    (0.5*(self.beta-self._base_beta)*np.log(kk) - np.log(qq_ratio)) * np.square(alpha)/qq
                 )
         #else:
         #    raise ValueError('Cannot handle this vol_beta different from beta')
 
-        order1 = term20 + term11 + term02
-
         ## Return
         if self.approx_order == 0:
             vol = 1.0
         else:
-            vol = 1.0 + order1*texp
+            vol = 1.0 + texp*(term20 + term11 + term02)
 
         pre_fac = alpha * np.power(fwd, vol_betac) * qq_ratio
         vol *= pre_fac*hh
         return vol
 
     def mass_zero(self, spot, texp, log=False):
-        assert(self.vol_beta is None)
+        assert(self._base_beta is None)
         vol_cev = self.vol_for_price(0.0, spot, texp)
         cev_m = cev.Cev(sigma=vol_cev, beta=self.beta)
         mass = cev_m.mass_zero(spot, texp, log=log)
-        #print(vol_cev, mass)
         return mass
 
     def mass_zero_t0(self, spot, texp):
@@ -406,7 +504,7 @@ class SabrChoiWu2021P(SabrChoiWu2021H, smile.MassZeroABC):
         # fwd, spot, sigma may be either scalar or np.array.
         # texp, vov, rho, beta should be scholar values
 
-        vol_beta = self.beta if self.vol_beta is None else self.vol_beta
+        vol_beta = self._base_beta or self.beta
         vol_betac = 1.0 - vol_beta
 
         fwd, _, _ = self._fwd_factor(spot, texp)
@@ -423,7 +521,7 @@ class SabrChoiWu2021P(SabrChoiWu2021H, smile.MassZeroABC):
 
         # qq_ratio = qq_vol_beta / qq_beta
         tmp = self._int_inv_locvol(kk, self.beta)
-        qq_ratio = 1.0 if self.vol_beta is None else self._int_inv_locvol(kk, vol_beta) / tmp
+        qq_ratio = 1.0 if self._base_beta is None else self._int_inv_locvol(kk, vol_beta) / tmp
 
         zz = vov_over_alpha_safe * tmp * (kk - 1.0)  # zeta = (vov/sigma0) q
         hh = self._hh(zz, self.rho)
@@ -468,7 +566,11 @@ class SabrChoiWu2021P(SabrChoiWu2021H, smile.MassZeroABC):
             gg_diff = self.rho*self.beta/(rhoc*betac) * (gg2 - gg1)
 
         zz2_safe = np.maximum(np.square(zz), np.finfo(float).eps)
-        tmp = 0 if self.vol_beta is None else 0.5*(self.beta-self.vol_beta)*np.log(kk) - np.log(qq_ratio)
+
+        if np.isclose(self.beta, vol_beta):
+            tmp = 0.0
+        else:
+            tmp = 0.5*(self.beta-self._base_beta)*np.log(kk) - np.log(qq_ratio)
 
         """
         with np.errstate(divide='ignore', invalid='ignore'):
@@ -476,7 +578,7 @@ class SabrChoiWu2021P(SabrChoiWu2021H, smile.MassZeroABC):
             term20 = np.where(
                 np.fabs(qq) < 1e-6,
                 (np.square(betac) - np.square(vol_betac)) / 24 * np.square(alpha),
-                (0.5 * (self.beta - self.vol_beta) * np.log(kk) - np.log(qq_ratio)) * np.square(alpha) / qq
+                (0.5 * (self.beta - self._base_beta) * np.log(kk) - np.log(qq_ratio)) * np.square(alpha) / qq
             )
         """
         order1 = np.square(self.vov*hh)/zz2_safe *\
